// $Id$
// A combination of the HAH basenode code and the generalized decoder and relay for 433 MHz OOK signals by
// 2010-04-11 <jc@wippler.nl> http://opensource.org/licenses/mit-license.php
// Thanks to Brett England, Derek kennedy
// 02-04-2014 Mark Baldwin

#include <JeeLib.h>
#include <util/crc16.h>
#include <util/parity.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>

// Do we want this HAHCentral to also be a RF receiver.
#define CENTRAL_RF

#ifdef CENTRAL_RF
  #include <RFRX.h>
  #include "decoders.h"
  #define DEBUG 1     // set to 1 to also report results on the serial port
  //#define DEBUG_LED // define as pin 1..19 to blink LED on each pin change
  #define PIN_433 17    // AIO4 = 433 MHz receiver
#endif


// Hardcode our node ID so it won't ever change once flashed.
#define NODE_ID 1

static unsigned long now () {
  // FIXME 49-day overflow
  return millis() / 1000;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// RF12 configuration setup code

typedef struct {
  byte nodeId;
  byte group;
  char msg[RF12_EEPROM_SIZE-4];
  word crc;
} 
RF12Config;

static RF12Config config;

static char cmd;
static byte value, stack[RF12_MAXDATA], top, sendLen, dest, quiet=1;
static byte testbuf[RF12_MAXDATA], testCounter;

static void addCh (char* msg, char c) {
  byte n = strlen(msg);
  msg[n] = c;
}

static void addInt (char* msg, word v) {
  if (v >= 10)
    addInt(msg, v / 10);
  addCh(msg, '0' + v % 10);
}
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

char helpText1[] PROGMEM = 
"\n"
"Available commands:" "\n"
"  ...,<nn> a - send data packet to node <nn>, with ack" "\n"
"  ...,<nn> s - send data packet to node <nn>, no ack" "\n"
"  <n> q      - set quiet mode (1 = don't report bad packets)" "\n"
;

static void showString (PGM_P s) {
  for (;;) {
    char c = pgm_read_byte(s++);
    if (c == 0)
      break;
    if (c == '\n')
      Serial.print('\r');
    Serial.print(c);
  }
}

static void showHelp () {
  showString(helpText1);
}

static void handleInput (char c) {
  if ('0' <= c && c <= '9')
    value = 10 * value + c - '0';
  else if (c == ',') {
    if (top < sizeof stack)
      stack[top++] = value;
    value = 0;
  } 
  else if ('a' <= c && c <='z') {
    Serial.print("> ");
    Serial.print((int) value);
    Serial.println(c);
    switch (c) {
    default:
      showHelp();
      break;
    case 'a': // send packet to node ID N, request an ack
    case 's': // send packet to node ID N, no ack
      cmd = c;
      sendLen = top;
      dest = value;
      memcpy(testbuf, stack, top);
      break;
    case 'q': // turn quiet mode on or off (don't report bad packets)
      quiet = value;
      break;
    }
    value = top = 0;
    memset(stack, 0, sizeof stack);   
  } 
  else if (c > ' ')
    showHelp();
}

#ifdef CENTRAL_RF
  #if PIN_433
      OregonDecoder orsc;
      CrestaDecoder cres;
      KakuDecoder kaku;
      KakuADecoder kakuA; //WvD
      XrfDecoder xrf;
      HezDecoder hez;
      FlamingoDecoder fmgo;
      SmokeDecoder smk;
      ByronbellDecoder byr;
      ElroDecoder elro;
      
      DecoderInfo di_433[] = {
          { 5, "ORSC", &orsc },
          { 6, "CRES", &cres },
          { 7, "BBSB", &kaku },
          { 8, "XRF", &xrf },
          { 9, "HE", &hez },
          { 10, "ELRO", &elro },
          { 11, "FMGO", &fmgo },
          { 12, "SMK", &smk },
          { 13, "BYR", &byr },
          { 14, "BBSBA", &kakuA },
          { -1, 0, 0 }
      };
      
      // State to track pulse durations measured in the interrupt code
      volatile word pulse_433;
      word last_433; // never accessed outside ISR's
      
      #if PIN_433 >= 14
      #define VECT PCINT1_vect
      #elif PIN_433 >= 8
      #define VECT PCINT0_vect
      #else
      #define VECT PCINT2_vect
      #endif
      
      ISR(VECT) {
          word now = micros();
          pulse_433 = now - last_433;
          last_433 = now;
      }
  #endif

// Outgoing data buffer for RF12
byte packetBuffer [RF12_MAXDATA], packetFill;

// Timer to only relay packets up to 10x per second, even if more come in.
MilliTimer sendTimer;

static void setupPinChangeInterrupt () {
#if PIN_433
    pinMode(PIN_433, INPUT);
    digitalWrite(PIN_433, 1);   // pull-up

    // interrupt on pin change
    #if PIN_433 >= 14
        bitSet(PCMSK1, PIN_433 - 14);
        bitSet(PCICR, PCIE1);
    #elif PIN_433 >= 8
        bitSet(PCMSK0, PIN_433 - 8);
        bitSet(PCICR, PCIE0);
    #else
        bitSet(PCMSK2, PIN_433);
        bitSet(PCICR, PCIE2);
    #endif
#endif
}

// Append a new data item to the outgoing packet buffer (if there is room
static void addToBuffer (byte code, const char* name, const byte* buf, byte len) {
#if DEBUG
    Serial.print(name);
    for (byte i = 0; i < len; ++i) {
        Serial.print(' ');
        Serial.print((int) buf[i]);
    }
    Serial.println();
#endif

    if (packetFill + len < sizeof packetBuffer) {
        packetBuffer[packetFill++] = code + (len << 4);
        memcpy(packetBuffer + packetFill, buf, len);
        packetFill += len;
    } else {
#if DEBUG
        Serial.print(" dropped: ");
        Serial.print(name);
        Serial.print(", ");
        Serial.print((int) len);
        Serial.println(" bytes");
#endif        
    }
}

static void addDecodedData (DecoderInfo& di) {
    byte size;
    const byte* data = di.decoder->getData(size);
    addToBuffer(di.typecode, di.name, data, size);
    di.decoder->resetDecoder();
}

// Check for a new pulse and run the corresponding decoders for it
static void runPulseDecoders (DecoderInfo* pdi, volatile word& pulse) {
    // get next pulse with and reset it - need to protect against interrupts
    cli();
    word p = pulse;
    pulse = 0;
    sei();

    // if we had a pulse, go through each of the decoders
    if (p != 0) { 
#if DEBUG_LED
        digitalWrite(DEBUG_LED, 1);
#endif
        while (pdi->typecode >= 0) {
            if (pdi->decoder->nextPulse(p))
                addDecodedData(*pdi);
            ++pdi;
        }
#if DEBUG_LED
        digitalWrite(DEBUG_LED, 0);
#endif
    }
}
#endif

void setup() {
  config.nodeId = NODE_ID;
  config.group = 212;
#if DEBUG
  Serial.begin(57600);
  Serial.print("\n[HAHCentral.1.1]");
  Serial.print(config.nodeId,DEC);
  Serial.print(" g");
  Serial.print(config.group,DEC);
  Serial.println(" @ 868Mhz");
#endif
  // RF12_868MHZ, RF12_915MHZ, RF12_433MHZ
  rf12_initialize(config.nodeId, RF12_868MHZ, config.group);
  
#ifdef CENTRAL_RF
  #if PIN_433
      setupPinChangeInterrupt();
      #if DEBUG
        Serial.println("433 Pin Change interupt setup");
      #endif
  #endif
#endif
}

void loop() {
  if (Serial.available())
    handleInput(Serial.read());

  if (rf12_recvDone()) {
    byte n = rf12_len;
    if (rf12_crc == 0) {
      Serial.print("OK");
    } 
    else {
      if (quiet)
        return;
      Serial.print(" ?");
      if (n > 20) // print at most 20 bytes if crc is wrong
        n = 20;
    }
    if (config.group == 0) {
      Serial.print("G ");
      Serial.print((int) rf12_grp);
    }
    Serial.print(' ');
    Serial.print((int) rf12_hdr);
    for (byte i = 0; i < n; ++i) {
      Serial.print(' ');
      Serial.print((int) rf12_data[i]);
    }
    Serial.println();

    if (rf12_crc == 0) {
      if (RF12_WANTS_ACK) {
        Serial.println(" -> ack");
        rf12_sendStart(RF12_ACK_REPLY, 0, 0);
      }
    }
  }

  if (cmd && rf12_canSend()) {
    Serial.print(" -> ");
    Serial.print((int) sendLen);
    Serial.println(" b");
    byte header = cmd == 'a' ? RF12_HDR_ACK : 0;
    if (dest)
      header |= RF12_HDR_DST | dest;
    rf12_sendStart(header, testbuf, sendLen);
    cmd = 0;
  }  
#ifdef CENTRAL_RF
  #if PIN_433
    runPulseDecoders(di_433, pulse_433);   
  #endif

    if (sendTimer.poll(100) && packetFill > 0) {
    packetFill = 0;
    }
#endif
}

